/*
 * PyJaMaS
 *
 * Copyright (c) 2006
 *	Peter Sabolcak
 * see ../README for more details
 */
 
 /** \file
 * Condition variable.
 *
 * Condition variable implementation as a part of synchronization primitives
 * used in kernel.
 */

#include "cond.h"
#include "intctl.h"
#include "debug.h"
#include "errno.h"
#include "thread.h"

#define SSD_COND_MAGIC 0xDEADBEEF

/**********************/
/* Condition variable */ 
/**********************/

/**
 * Condition variable initialization.
 *
 * This function sets \c cvar to init state which means that an empty list for
 * pending threads to be held in is created.
 *
 * \param cvar condition variable structure which should be initialized
 */
void cond_init ( struct cond *cvar )
{
	int_state_t int_state;

	save_and_disable_interrupts(int_state);

	#ifdef STATIC_STRUCT_DEBUG
	if (!KERNEL_PTR_VALID (tmr))
		panic ("Invalid condvar pointer (%p)",cvar);
	tmr->magic = SSD_COND_MAGIC;
	#endif

	list_init(&cvar->waiting);
	
	restore_interrupts(int_state);
}

/**
 * Condition variable destruction
 *
 * The function checks whether the condition variable can be destroyed. If it
 * can not be destroyd it means that there is still a pending lock on
 * condition variable and it calls \c panic().
 *
 * \param cvar condition variable structure which should be destroy
 */
void cond_destroy ( struct cond *cvar )
{
	int_state_t int_state;

	save_and_disable_interrupts(int_state);
	
	#ifdef STATIC_STRUCT_DEBUG
	if (!KERNEL_PTR_VALID (cvar))
		panic ("Invalid condvar pointer (%p)",cvar);
	if (cvar->magic != SSD_COND_MAGIC)
		panic ("Invalid condvar magic (%x)",cvar->magic);
	#endif

	//Test whether the condition variable is ready to be destroyed
	if(!list_is_empty(&cvar->waiting))
		panic("There are active locks in condition variable which is \
going to be killed.");

	restore_interrupts(int_state);
}

/**
 * Condition variable signal 
 *
 * The function wakes up the thread at the beginning of the waiting list. This
 * method (using FIFO list) should prevent starvation. This thread will be
 * removed from the list.
 *
 * \param cvar condition variable structure on which the
 *		thread should be awoken
 */
void cond_signal ( struct cond *cvar )
{
	struct thread_struct *thrd;
	struct list_entry *e;

	int_state_t int_state;

	save_and_disable_interrupts(int_state);

	#ifdef STATIC_STRUCT_DEBUG
	if (!KERNEL_PTR_VALID (cvar))
		panic ("Invalid condvar pointer (%p)",cvar);
	if (cvar->magic != SSD_COND_MAGIC)
		panic ("Invalid condvar magic (%x)",cvar->magic);
	#endif

	//For emtpy list just continue, nothing can be awaken
	if(list_is_empty(&cvar->waiting))
	{
		restore_interrupts(int_state);
		return;
	}

	e = list_first(&cvar->waiting);

	//If we get here it mean that everything is ok and we can wake up
 	//so we obtain a ptr to the thread struct
	thrd = LIST_DATA_PTR (e,  struct thread_struct);

	list_remove (e);

	//Wake up of first thread in queue
	thread_revive (thrd->id); 

	restore_interrupts(int_state);
}

/**
 * Condition variable broadcast
 *
 * The function wakes up all threads which are in the pending list on a
 * condition variable. If no threads are present in the queue it just returns.
 *
 * \param cvar condition variable structure on which all threads should
 *		be awoken
 */
void cond_broadcast ( struct cond *cvar )
{
	int_state_t int_state;

	save_and_disable_interrupts(int_state);

	#ifdef STATIC_STRUCT_DEBUG
	if (!KERNEL_PTR_VALID (cvar))
		panic ("Invalid condvar pointer (%p)",cvar);
	if (cvar->magic != SSD_COND_MAGIC)
		panic ("Invalid condvar magic (%x)",cvar->magic);
	#endif

	//For empty list just continue, nothing can be wake up
	if(list_is_empty(&cvar->waiting))
	{
		restore_interrupts(int_state);
		return;
	}

	//Wake up for all waiting threads
	_list_wakeup(&cvar->waiting);

	restore_interrupts(int_state);
}

/**
 * Condition variable wait 
 *
 * The function automatically unlocks the \c mtx mutex and suspends the calling
 * thread on condition variable. After it wakes up it tries to lock the mutex
 * again and returns.
 *
 * \param cvar condition variable structure on which we want to wait
 * \param mtx mutex which protects the access to decision variable
 */
void cond_wait (struct cond * cvar, struct mutex * mtx)
{
	int_state_t int_state;

	save_and_disable_interrupts(int_state);
	
	#ifdef STATIC_STRUCT_DEBUG
	if (!KERNEL_PTR_VALID (cvar))
		panic ("Invalid condvar pointer (%p)",cvar);
	if (cvar->magic != SSD_COND_MAGIC)
		panic ("Invalid condvar magic (%x)",cvar->magic);
	#endif

	mutex_unlock(mtx);

	list_insert_back(&cvar->waiting, 
			&(current_thread->sync_queue),
			(void*)current_thread);

	thread_wait();

	restore_interrupts(int_state);

	mutex_lock(mtx);
}

/**
 * Condition variable wait with timeout 
 *
 * The function automatically unlocks \c mtx mutex and suspends the calling
 * thread on a condition variable for some time. If the thread is not awaken
 * during this period then \c ETIMEDOUT is returned, otherwise EOK is returned.
 * If \c usec is 0, the mutex is unlocked and the function calls
 * \c thread_yield() so it loses its running status.
 *
 * \param cvar condition variable structure on which we want to wait
 * \param mtx mutex which protects the access to decision variable
 * \param usec how long we should sleep, in miroseconds
 */
int cond_wait_timeout (struct cond * cvar, struct mutex * mtx, 
			const unsigned int usec)
{
	int_state_t int_state;
	unsigned int usec_left;

	usec_left = usec;

	save_and_disable_interrupts(int_state);
	
	#ifdef STATIC_STRUCT_DEBUG
	if (!KERNEL_PTR_VALID (cvar))
		panic ("Invalid condvar pointer (%p)",cvar);
	if (cvar->magic != SSD_COND_MAGIC)
		panic ("Invalid condvar magic (%x)",cvar->magic);
	#endif

	mutex_unlock(mtx);

	if( usec_left == 0 )
	{
		//Schedule and then try to lock mutex
		thread_yield();

	}
	else
	{	
		//Insert current thread to the waiting queue
		list_insert_back(&cvar->waiting,
			&(current_thread->sync_queue),
			(void*)current_thread);
			
		//Wait for timeout or a signal/broadcast. If, after falling
		//out of thread_wait_usec, we're still in the queue,
		//then it's a false revival so sleep again. (shouldn't happen)
		while( usec_left > 0 &&
			list_entry_is_used (&current_thread->sync_queue) )
		{
			#ifdef DEBUG
			printk("SLEEP\n");
			#endif

			usec_left = thread_wait_usec(usec_left);
		}
		//If we are still in queue that means that we timeouted
		if( list_entry_is_used (&current_thread->sync_queue) )
		{	//We have not been awaken by any signal or broadcast
			list_remove(&current_thread->sync_queue);

			restore_interrupts(int_state);

			mutex_lock(mtx);

			return ETIMEDOUT;
		}
	}

	restore_interrupts(int_state);

	mutex_lock(mtx);

	return EOK; 
}
