/*
 * PyJaMaS
 *
 * Copyright (c) 2006
 *	Jiri Svoboda
 * see ../README for more details
 */
 
 /** \file
 * Linked Lists.
 *
 * Statically allocated doubly-linked lists with a head.
 */

#include "list.h"
#include "debug.h"
#include "thread.h"

//#define DEBUG

/**
 * Initialise a list structure.
 */
void list_init (struct list *l)
{
	l->head.prev = l->head.next = &l->head;
}

/**
 * Insert entry at the beginning of a list.
 */
void list_insert_front (struct list *l, struct list_entry *e, void *data)
{
	#ifdef DEBUG
	if (!l) panic("NULL list argument\n");
	#endif

	e->next = l->head.next;
	e->prev = &l->head;
	e->data = data;
	
	l->head.next->prev = e;
	l->head.next = e;
}

/**
 * Insert entry at the end of a list.
 */
void list_insert_back (struct list *l, struct list_entry *e, void *data)
{
	#ifdef DEBUG
	if (!l) panic("NULL list argument\n");
	#endif

	e->next = &l->head;
	e->prev = l->head.prev;

	e->data = data;
	
	l->head.prev->next = e;
	l->head.prev = e;
	
}

/**
 * Remove an entry from a list.
 */
void list_remove (struct list_entry *e)
{
	#ifdef DEBUG
	if (!l) panic("NULL list argument\n");
	#endif

	e->next->prev = e->prev;
	e->prev->next = e->next;
	
	e->next = NULL;
	e->prev = NULL;
}

/**
 * Remove all entries from a list.
 */
void list_remove_all (struct list *l)
{
	struct list_entry *e;

	#ifdef DEBUG
	if (!l) panic("NULL list argument\n");
	#endif
	
	e = list_first (l);
	while(!list_is_head (l, e)) {
	    list_remove (e);
	    e = list_first (l);
	}
}

/**
 * Move one entry from the beginning of a list to the end.
 *
 * The head advances forword, i.e. the list entries rotate backward
 * with respect to the head.
 */
void list_rotate_backward (struct list *l)
{
	struct list_entry *e;

	#ifdef DEBUG
	if (!l) panic("NULL list argument\n");
	#endif
	
	e = list_first (l);
	list_remove (e);
	list_insert_back (l, e, e->data);
}

/**
 * Return the first entry of a list.
 */
struct list_entry *list_first (struct list *l)
{
	#ifdef DEBUG
	if (!l) panic("NULL list argument\n");
	#endif

	return l->head.next;
}

/**
 * Return the successor of an entry.
 */
struct list_entry *list_next (struct list_entry *e)
{
	return e->next;
}

/**
 * Return nonzero if a list is empty.
 */
bool_t list_is_empty (struct list *l)
{
	#ifdef DEBUG
	if (!l) panic("NULL list argument\n");
	#endif

	return l->head.next == &(l->head);
}

/**
 * Return nonzero if an entry is the list head.
 *
 * Note that the list head does not contain any data, so
 * this is used to detect we have reached the end of a list.
 */
bool_t list_is_head (struct list *l, struct list_entry *e)
{
	#ifdef DEBUG
	if (!l) panic("NULL list argument\n");
	#endif

	return e == &(l->head);
}

/**
 * Initialise the fields of an entry with zeros.
 *
 * Using this is a good thing (tm) and makes it possible
 * to check whether the entry is being used by checking
 * its contents.
 */
void list_entry_init (struct list_entry *entry)
{
	entry->prev = entry->next = NULL;
	entry->data = NULL;
}

/**
 * Return nonzero if an entry is being used.
 */
bool_t list_entry_is_used (struct list_entry *entry)
{
	return entry->prev != NULL;
}

/**
 * Return the number of elements in a list.
 */
int list_count (struct list *l)
{
	struct list_entry *ent;
	int count;

	#ifdef DEBUG
	if (!l) panic("NULL list argument\n");
	#endif
	
	count = 0;
	ent = list_first(l);

	while(!list_is_head(l, ent))
	{
		ent = list_next(ent);
		count++;
	}
	return count;
}
