/*
 * Kalisto
 *
 * Copyright (c) 2001-2005
 * 	Distributed Systems Research Group
 * 	Department of Software Engineering
 * 	Faculty of Mathematics and Physics
 * 	Charles University, Czech Republic
 * see ../README for more details
 */
 
/** \mainpage
 *
 * PyJaMaS is the first keystone of an operating system kernel.
 * It is based on Kalisto and adds several significant features,
 * such as timers, kernel thread support and synchronization primitives.
 *
 * Currently PyJaMaS runs on a simulated MIPS machine within the msim
 * emulator. The kernel uses preemptive multitasking, SMP is not
 * supported, however.
 *
 * \section fixes Notable Kalisto Code Improvements
 *
 * - A bug in sys.S was fixed where the switch_cpu_context function
 *   would not restore the A0 register.
 * - A stack frame structure \c struct \c stack_frame was defined
 *   in stackframe.h to replace offset.h in C code. Now offset.h
 *   is only used in assembly code and is generated automatically
 *   by the genoffset tool (see genoffset.c and stackframe.h).
 *   The types of \c kernel_stack_frame and \c kernel_stack_top in 
 *   \c struct \c thread_struct have been changed to \c stack_frame pointers
 *   and renamed to more appropriate names \c exception_context and
 *   \c task_switch_context. This also fixed the 'dereferencing
 *   type-punned pointer will break strict aliasing' warning.
 * - The special thread list type has been removed and general lists
 *   have been introduced instead.
 * - malloc() and free() have been updated to work with both with
 *   enabled or disabled interrupts interrupts and alloc_mem() and
 *   free_mem() have been made static.
 *
 * \section contents Contents
 *
 * You can read about how different parts of PyJaMaS work in the
 * detailed descriptions of relevant modules. Here's a list of the
 * most important modules sorted by categories. See the Files tab
 * for a list of all files.
 *
 * \subsection scheduling Threads, Timers, Scheduling
 *	- thread.c
 *	- timer.c
 *	- sched.c
 *
 * \subsection synchronization Synchronization
 *	- mutex.c
 *	- sem.c
 *	- cond.c
 *	- rwlock.c
 *
 * \subsection data_structures Data Structures
 *	- list.c
 *	- heap.c
 *	- hashtable.c
 *
 * \subsection io Input / Output
 *	- printer.c
 *	- keyboard.c
 *
 * \subsection exc_int Exceptions, Interrupts
 *	- exc.c
 *	- int.c
 *
 * \section credits PyJaMaS Credits
 *	- Peter Sabolcak <bondoer@matfyz.cz>
 *		synchronization primitives, I/O
 *	- Jiri Svoboda <chip@matfyz.cz>
 *		threads, timers, lists, heaps, docs
 * 	- Michal Zofka <uzofka@matfyz.cz>
 *		hash tables, documentation
 *
 *
 * \section Kalisto Credits
 *
 *	- Viliam Holub <holub@nenya.ms.mff.cuni.cz>
 *		main design
 *	- Petr Tuma <petr.tuma@mff.cuni.cz>
 *		notes
 *	- Lubomir Bulej <lubomir.bulej@mff.cuni.cz>
 *	- Tomas Bures <bures@nenya.ms.mff.cuni.cz>
 *		notes, comments
 *
 * \section Final Word
 *
 * Writing PyJaMaS has been a lot of work, but a lot of fun, too.
 * We do hope you will find this piece of code useful.
 * 
 * 10. 11. 2006
 *
 * Any comments and suggestions are welcome.
 * The team
 */

#include "malloc.h"
#include "thread.h"
#include "sched.h"
#include "tlb.h"
#include "sys.h"
#include "timer.h"
#include "printer.h"
#include "keyboard.h"

#include "example.h"

#include "simdebug.h"

void assignment_test (void);

static void _initial_thread_proc (void *data)
{
	(void)data;
	
	assignment_test ();
//	example ();

	___halt ();
}

/*
 * This is the entry point to C code. The "start" function is called from
 * the basic initialization routine in "sys.S". We have to create and
 * initialize the main data structures.
 *
 * The interrupts are (still) disabled.
 */
void start (void)
{
	struct thread_struct *t;

	/*
	 * Initialize environment.
	 */
	print ("Initializing PyJaMaS... ");

	/* Set TLB to a defined state. */
	init_tlb ();

	/* Build memory structures so we can use malloc and free. */
	init_memory ();

	/* Init threads and create an initial thread */
	t = threads_init (_initial_thread_proc, NULL);

	/* Init scheduler and add our new thread. */
	sched_init (t);

	/* Init the timer module. */
	timers_init ();
	
	/* Init keyboard module. */
	init_keyboard ();

	/* Create a thread to service timers and join detached threads */
	thread_init_service_thread ();

	print ("done\n\n");

	/*
	 * All necessary data structures have been initialized and we
	 * are ready to do some work.
	 */

	/*
	 * Enter the initial thread.
	 *
	 * Note: our current thread of execution is not stored in the 
	 * thread/scheduler structures. Thus the call to context_switch()
	 * will switch into the new thread (passed as argument to sched_init())
	 * and this thread context will be
	 * forgotten.
	 */
	context_switch (t);

	/* unreachable */
}
