/*
 * PyJaMaS
 *
 * Copyright (c) 2006
 *	Jiri Svoboda
 *	portions from Kalisto
 * see ../README for more details
 */

 /** \file
 * Simple round-robin thread scheduler with suspending.
 */

#include "sched.h"

#include "thread.h"
#include "timer.h"

#include "simdebug.h"
#include "debug.h"
#include "printer.h"


/**
 * Default amount of time a thread is alloted.
 */
#define THR_DEFAULT_QUANTUM	2024

/**
 * Randomization of a thread quantum to get more
 * interesting behavior in the example.
 */
#define THR_RANDOM_QUANTUM	(rnd () & 0x3ff)

/**
 * A list of ready-to-run threads. This serves as a round-robin queue -- the
 * thread at the head will be activated by a \c schedule() call.
 */
static struct list runnable_list;

/**
 * List of sleeping threads. These threads are waiting on a synchronization
 * primitive, in a sleep/usleep/join function, waiting to be joined etc.
 * or have been suspended by a call to \c thread_suspend()
 */
static struct list sleeping_list;

/**
 * Init the scheduler module.
 *
 * Mark the thread \a t as first runnable.
 */
void sched_init (struct thread_struct *t)
{
	/* Init thread lists */
	list_init (&runnable_list);
	list_init (&sleeping_list);

	/* Make the first thread runnable */
	sched_add_runnable_thread (t);

	/* Set up the first task switch */
	timer_set_task_switch (4024);
}


/**
 * Mark the thread as ready to run and add it to the ready-to-run list.
 */
void sched_add_runnable_thread (struct thread_struct *t)
{
	t->state = TS_READY_TO_RUN;
	list_insert_back (&runnable_list, &t->scheduling, (void *)t);
}

/**
 * Move the thread from runnable list to sleeping list.
 */
void sched_sleep_thread (struct thread_struct *t)
{
	assert (t->state == TS_READY_TO_RUN || t->state == TS_RUN);
	t->state = TS_SLEEPING;
	list_remove (&t->scheduling);
	list_insert_back (&sleeping_list, &t->scheduling, (void *)t);
}

/**
 * Move the thread from sleeping list to runnable list.
 */
void sched_unsleep_thread (struct thread_struct *t)
{
	assert (t->state == TS_SLEEPING);
	t->state = TS_READY_TO_RUN;
	list_remove (&t->scheduling);
	list_insert_back (&runnable_list, &t->scheduling, (void *)t);
}

/**
 * Remove the thread from either runnable list or sleeping list.
 */
void sched_remove_thread (struct thread_struct *t)
{
	list_remove (&t->scheduling);
}

/*
 * Primitive random number generator.
 */

static unsigned rseed = 1234567;

/**
 * Return a random unsigned number.
 */
static unsigned rnd ()
{
	rseed = (rseed * 873511) % 22348977 + 7;
	return rseed >> 8;
}


/**
 * Schedule the next thread.
 *
 * Interrupts must to be disabled.
 */
void schedule (void)
{
	struct thread_struct *next_thread;

	/* Take the first thread in the list and make it the last... */
	next_thread = LIST_DATA_PTR(
		list_first (&runnable_list), struct thread_struct);
	list_rotate_backward (&runnable_list);
	
//	printk("[sched %d]\n",next_thread->id);

	/*
	 * Set a new time to switch.
	 */
	timer_set_task_switch(
		THR_DEFAULT_QUANTUM + THR_RANDOM_QUANTUM);

	/* Switch context to the newly selected thread. */
	if (current_thread != next_thread)
		context_switch (next_thread);
}
