/*
 * PyJaMaS
 *
 * Copyright (c) 2006
 *	Jiri Svoboda
 *	portions from Kalisto
 * see ../README for more details
 */
 
 /** \file
 * Kernel Threads.
 *
 * \section thread_creation Creation
 *
 * A new thread is allocated an id, a kernel stack, and its initial
 * context is created. It is inserted into scheduler structures
 * and starts executing. All threads begin by entering the _thread_launch()
 * function. Its purpose is to wait for someone to clean the thread
 * up, after it has finished execution.
 *
 * \section thread_states States
 *
 * Threads can be in various states:
 *	- \i running - currently running on the CPU
 *	- \i ready-to-run - waiting to get the CPU
 *	- \i waiting - waiting on a synchronization primitive or join()
 *          or waiting to be joined
 *	- \i suspended - as a result of thread_suspend()
 *
 * Threads that are either waiting or suspended are called sleeping,
 * and are not scheduled for execution. The \i waiting and \i suspended states
 * are similar, but the first is for internal use, whereas the second one
 * is for public use throughout the kernel.
 *
 * \section thread_killing Premature Destruction
 *
 * A thread may be killed by another thread using thread_kill().
 * However, as of yet there are limitations to when a thread
 * can be killed. Simply put, threads waiting on synchronization primitives
 * must not be killed.
 *
 * \section thread_termination Termination
 *
 * When a thread finishes execution, it passively waits for another thread
 * to thread_join() it. If there already is another thread waiting to join it,
 * it is woken up.
 *
 * \section detached_threads Detached Threads
 *
 * Since freeing the current thread's memory would be rather cumbersome,
 * a finished detached threads enlists itself to a 'finished_detached'
 * list of the service thread. The service thread periodically checks this
 * list and cleans up any finished detached threads.
 *
 * \section The 'service' thread
 *
 * The 'service' thread is is scheduled the same way as any other ordinary
 * thread. Since the scheduling is round-robin, it is executed every now and then.
 * Every time it is scheduled, this thread:
 *	- cleans up finished detached threads
 *	- services timers (enlist expired timers to their respective threads)
 *	- gives up the CPU
 * This ensures for each timer that, if its owner thread has not any other
 * timer handlers running, its handler begins executing no later than
 * two rounds after its planned expiration.
 *
 * Note also that the service thread is (naturally) preemptible.
 */

#include "thread.h"

#include "sys.h"
#include "intctl.h"
#include "malloc.h"
#include "list.h"
#include "sched.h"
#include "timer.h"
#include "errno.h"
#include "debug.h"
#include "stackframe.h"
#include "hashtable.h"

#include "printer.h"

/**
 * Hashtable to convert thread id to thread structure pointer.
 */
struct hash *thread_ident;

/**
 * Next id to be allocated
 */
thread_t next_id = 1;

/**
 * Kernel stack size.
 *
 * Keep it big enough - stack overflow often causes strange
 * behavior which is hard to find cause for.
 */
#define KERNEL_STACK_SIZE 4096

/*
 * The main CP0 status configuration for kernel threads. Switch to the kernel
 * mode, enable all interrupts and mark coprocessor 0 as usable (this is
 * not necessary because cp0 is available in the kernel mode automatically).
 */
#define KERNEL_THREAD_STATUS	\
(				\
	cp0_status_kx_mask 	\
	| cp0_status_ie_mask 	\
	| cp0_status_im_mask	\
	| cp0_status_cu0_mask	\
)

/**
 * A pointer to the currently running thread.
 */
struct thread_struct *current_thread;


/**
 * Detached threads are cleaned-up by the service thread
 */
struct list finished_detached_threads;

/**
 * Initializes the context in a thread.
 *
 * The thread's stack must have been already allocated.
 */
static void _init_thread_context (struct thread_struct *t,
			void (*start) (void (*start) (void *data),void *),
			int arg1, int arg2, int status)
{
	struct stack_frame *sf;
	
	/*
	 * This is set when an exception is raised
	 */
	t->exception_context = 0;
	
	/*
	 * Allocate a stack frame on top of the thread's stack
	 */
	t->task_switch_context = (struct stack_frame *)
		 &t->kernel_stack [
		KERNEL_STACK_SIZE / sizeof (int) -
		EX_STACK_FRAME / sizeof (int) - 1];
		
	sf = t->task_switch_context;
	
	sf->ra = (int) start;
	sf->gp = (int) OM_KSEG1;
	sf->a0 = arg1;
	sf->a1 = arg2;
	sf->status = status;
}

/**
 * Used as entry point for new threads.
 *
 * Runs the thread's start routine. After start() returns, it waits
 * to be joined by another thread, or, if it has been detached,
 * by the service thread.
 */
static void _thread_launch (void (*start) (void *data), void *data)
{
	int_state_t int_state;
	
	start (data);
	
	save_and_disable_interrupts (int_state);
	current_thread->is_finished = 1;
	
	if (current_thread->is_waited_for)
		thread_revive (current_thread->waited_for_by);

	if (current_thread->is_detached) {
		list_insert_back (&finished_detached_threads,
			    &current_thread->finished_detached,
			    (void *)current_thread->id);
		/* Add this thread to the service thread's finished_detached list */
	}
	
	/*
	 * Wait for somebody to thread_join() us
	 */
	while (1)
	    thread_wait ();
	    
	/* Unreachable */
}

/**
 * Release a thread id.
 */
static void _free_thread_id (thread_t thr_id)
{
	hashtable_delete (thread_ident, thr_id);
}


/**
 * Create a new thread.
 *
 * Returns a pointer to the new thread structure. \a start is a function
 * which is used as an entry point. The thread NOT inserted into scheduler
 * structures. Must be run with interrupts disabled.
 *
 * Returns 0 if the thread could not be initialized.
 */
static struct thread_struct *_new_thread (void (*start) (void *data),
					void *data, int status)
{
	struct thread_struct *t;

	/*
	 * Allocate memory for a thread control structure and stack.
	 */
	t = malloc (sizeof (struct thread_struct));
	if (!t)
		return 0;

	t->kernel_stack = malloc (KERNEL_STACK_SIZE);
	if (!t->kernel_stack) {
		free (t);
		return 0;
	}

	/* The thread is not a part of scheduler structures yet */
	t->state = TS_DISABLED;
	
	/* Normal code will be ready to run */
	t->rc_state = TRCS_READY;
	
	/* No timer handler code running yet */
	t->tc_state = TTCS_INACTIVE;
	
	/*
	 * Initialize thread state, stack and context.
	 */
	_init_thread_context (t, _thread_launch, (int)start, (int)data, status);

	/*
	 * Allocate id, insert to hashtable.
	 */
	do {
		t->id = next_id++;
	} while (hashtable_retrieve (thread_ident, t->id) != NULL);
	
	hashtable_insert (thread_ident, t->id, (void *)t);

	t->is_waited_for = 0; t->waited_for_by = 0;
	t->is_finished = 0;
	t->is_detached = 0;
	
	list_entry_init (&t->scheduling);
	list_entry_init (&t->finished_detached);
	list_entry_init (&t->sync_queue);

	/*
	 * The thread starts with an empty list of pending timers
	 * and expired timers.
	 */
	list_init (&t->pending_timers);
	list_init (&t->expired_timers);

	return t;
}


/**
 * Remove a thread and free all its resources.
 *
 * Frees all the resources used by a thread and removes it from all
 * kernel structures.
 *
 * For situations when it's safe to kill/clean up a thread, see thread_kill().
 *
 * Must be run with interrupts disabled
 */
static void _cleanup_thread (struct thread_struct *ts) {

	sched_remove_thread (ts);
	
	/*
	 * Destroy any pending timers. Note that handlers can still
	 * be in the middle of execution.
	 */
	timer_destroy_pending_timers (ts);
	
	/*
	 * Remove from the finished_detached_threads list
	 */
	if(ts->is_finished && ts->is_detached)
		list_remove (&ts->finished_detached);

	/*
	 * Free the stack, thread_struct and id
	 */
	_free_thread_id (ts->id);
	free (ts->kernel_stack);
	free (ts);
}

/**
 * Init thread module and create initial thread.
 * 
 * Later the first task switch will switch into the initial thread.
 */
struct thread_struct *threads_init (void (*start) (void *data), void *data)
{
	struct thread_struct *t;

	thread_ident = hashtable_init();

	t = _new_thread (start, data, KERNEL_THREAD_STATUS);
	if (!t)
		panic("Failed to create initial thread");
	
	current_thread = t;

	list_init (&finished_detached_threads);

	return t;
}

/**
 * Clean up detached threads that have finished execution.
 */
static void _cleanup_finished_detached (void)
{
	struct list_entry *thr_entry, *next_entry;
	struct thread_struct *ts;
	thread_t thr;	
	int_state_t int_state;
	
	save_and_disable_interrupts (int_state);

	thr_entry = list_first (&finished_detached_threads);
	while (!list_is_head (&finished_detached_threads, thr_entry)) {
	
		next_entry = list_next (thr_entry);
		thr = LIST_DATA (thr_entry, thread_t);
		
		/* This also removes thr from the list automatically */
		ts = thread_get_struct (thr);
		if (ts)
			_cleanup_thread (ts);
		
		thr_entry = next_entry;
	}
	
	restore_interrupts (int_state);
}

/**
 * Service thread entry point.
 *
 * The service thread periodically services timers and cleans up
 * finished detached threads.
 */
static void _service_thread_proc (void *unused)
{	
	(void)unused;
	
	while (1) {
		_cleanup_finished_detached ();
		timer_service_timers ();
		
		/* Yes, this is THE case where using thread_yield is ok */
		thread_yield();
	}
	/* unreachable */
}

/**
 * Create the service thread.
 */
void thread_init_service_thread (void)
{
	thread_t id;
	int res;
	
	res = thread_create (&id, _service_thread_proc, NULL);
	
	if(res != EOK)
		panic("Failed to create service thread");
}

/**
 * Return b_true if current thread has expired timers.
 */
bool_t _have_expired_timer (void)
{
	int_state_t int_state;
	bool_t res;

	save_and_disable_interrupts (int_state);
	res = !list_is_empty (&current_thread->expired_timers);
	restore_interrupts (int_state);
	
	return res;
}

/**
 * Run timer handler code for the current thread.
 *
 * If there are expired timers for the current thread,
 * the thread enters timer handling mode and runs the handlers.
 */
static void _run_timer_handler_code (void) {
	if (_have_expired_timer ()) {
		current_thread->tc_state = TTCS_ACTIVE;
			
		/*
		 * Run timer handlers with interrupts
		 * enabled to make them preemptible.
		 */
		enable_interrupts ();
		timer_run_handlers ();
		disable_interrupts ();
			
		current_thread->tc_state = TTCS_INACTIVE;
	}
}

/**
 * Switch from current context to context of a given thread.
 *
 * Must be run with interrupts disabled.
 */
void context_switch (struct thread_struct *t)
{
	struct thread_struct *ct = current_thread;

	if (current_thread->state == TS_RUN)
		current_thread->state = TS_READY_TO_RUN;

	while (1) {
		/* Switch to the next thread */
		t->state = TS_RUN;
		current_thread = t;
		switch_cpu_context (&ct->task_switch_context, t->task_switch_context);
		
		/*
		 * Got the CPU back.
		 *
	         * From here on current_thread == ct again
		 */
		
		if (ct->tc_state == TTCS_INACTIVE) {
			/* Run any expired timer handlers */
			_run_timer_handler_code ();
		} else  {
			/* 
			 * A timer handler already running,
			 * break immediately.
			 */
			break;
		}
		
		/*
		 * If the thread is ready to run, return to
		 * regular code execution.
		 */
		if (ct->rc_state == TRCS_READY)
			break;
	
		/*
		 * Regular code is still waiting or suspended.
		 * So move the thread to sleeping list and loop.
		 */
		sched_sleep_thread (ct);
	}

	/* Back to regular or timer code */
}


/**
 * Return the thread_struct * associated with an id.
 *
 * Returns NULL if the id is (no longer) valid.
 *
 * Note that thread_struct * can only be considered valid as long as
 * interrupts are disabled.
 */
struct thread_struct *thread_get_struct (thread_t thr_id)
{
	struct thread_struct *ts;
	int_state_t int_state;

	save_and_disable_interrupts (int_state);
		
	ts = hashtable_retrieve (thread_ident, thr_id);
	
	restore_interrupts (int_state);
		
	return ts;
}

/**
 * Create a new runnable thread.
 *
 * The new thread is immediately prepared for execution.
 */
thread_t thread_create (thread_t *thread_ptr, void (*thread_start) (void *),
			void *data)
{
	struct thread_struct *thr;
	int_state_t int_state;

	save_and_disable_interrupts (int_state);
	
	thr = _new_thread (thread_start, data, KERNEL_THREAD_STATUS);
	if (!thr) {
		restore_interrupts (int_state);
		return ENOMEM;
	}

	*thread_ptr = thr->id;
	

	sched_add_runnable_thread (thr);
	

	return EOK;
}

/**
 * Return id of the current thread.
 */			
thread_t thread_get_current (void)
{
	return current_thread->id;
}

/**
 * Wait for a thread to finish execution.
 *
 * Blocks the current thread indefinitely, until \a thr finishes execution.
 */
int thread_join (thread_t thr)
{
	struct thread_struct *ts;
	int_state_t int_state;
	
	save_and_disable_interrupts (int_state);

	ts = thread_get_struct (thr);
	if (!ts || ts == current_thread) {
		restore_interrupts (int_state);
		return EINVAL;
	}

	if (ts->is_waited_for || ts->is_detached) {
		restore_interrupts (int_state);
		return EINVAL;
	}
	
	ts->is_waited_for = 1;
	ts->waited_for_by = current_thread->id;
	
	/*
	 * Wait for thread thr to finish. When thr finishes,
	 * it will revive us.
	 */
	while (!ts->is_finished)
	    thread_wait ();

	/* Remove the thread completely */
	_cleanup_thread (ts);

	restore_interrupts (int_state);

	return EOK;
}

/**
 * Wait for a thread to finish with timeout.
 *
 * Blocks the current thread until \a thr finishes execution,
 * or at least \a usec microseconds have passed.
 */
int thread_join_timeout (thread_t thr, const unsigned int usec)
{
	struct thread_struct *ts;
	int_state_t int_state;
	unsigned int usec_left;
	
	save_and_disable_interrupts (int_state);
	
	ts = thread_get_struct (thr);
	
	if (!ts || ts == current_thread) {
		restore_interrupts (int_state);
		return EINVAL;
	}

	if (ts->is_waited_for || ts->is_detached) {
		restore_interrupts (int_state);
		return EINVAL;
	}
	
	ts->is_waited_for = 1;
	ts->waited_for_by = current_thread->id;

	/*
	 * Loop until either no time left or ts has finished.
	 * The loop is just to be more robust.
	 * ts will revive us when it finishes
	 */		
	usec_left = usec;
	while (usec_left > 0 && !ts->is_finished) {
		usec_left = thread_wait_usec (usec_left);
	}

	ts->is_waited_for = 0;
	ts->waited_for_by = 0;
	
	if (!ts->is_finished) {
		restore_interrupts (int_state);
		return ETIMEDOUT;
	}

	/* Remove the thread completely */
	_cleanup_thread (ts);

	restore_interrupts (int_state);
	
	return EOK;
}

/**
 * Detach the current thread.
 *
 * A detached thread cannot be waited for with thread_join()
 * and its resources are automatically freed by the service thread.
 */
int thread_detach (thread_t thr)
{
	struct thread_struct *ts;
	int_state_t int_state;
	
	save_and_disable_interrupts (int_state);

	ts = thread_get_struct (thr);
	if (!ts) {
		restore_interrupts (int_state);
		return EINVAL;
	}

	if (ts->is_finished) {
		restore_interrupts (int_state);
		return EINVAL;
	}
	
	if (ts->is_waited_for || ts->is_detached) {
		restore_interrupts (int_state);
		return EINVAL;
	}

	ts->is_detached = 1;	

	restore_interrupts (int_state);
	
	return EOK;
}

/**
 * Suspend the thread for at least \a sec seconds.
 *
 * Must not be called from a timer handler.
 */
void thread_sleep (const unsigned int sec)
{
	unsigned int left;
	
	left = sec;
	
	/*
	 * Sleep for each second using thread_usleep()
	 */
	while (left > 0) {	
		thread_usleep (1000000);
		--left;
	}
}

/**
 * Timer handler: revive a thread from thread_wait_usec()
 *
 * This handler is executed with interrupts disabled.
 */
static void _revive_me (struct timer *tmr, void *thr) 
{
	(void)tmr;
	
	/* This can fail if the thread has been killed in the meantime */
	thread_revive ((thread_t)thr);
}

/**
 * Wait for at least \a usec microseconds, or for thread_revive().
 *
 * Makes the thread wait for at least usec microseconds, unless
 * the thread is revived earlier by another thread calling thread_revive()
 * thread_wait_usec returns the amount of remaining microseconds (or zero).
 *
 * This can be easily used for implementing passive-waiting for a condition
 * with a timeout. (Disabling interrupts is used for locking)
 *
 * Notes:
 *   - must not be called from a timer handler.
 *   - if 0 is returned, this does NOT necessarily mean that the thread
 *     hasn't been revived
 *   - thread_wakeup() has no effect on the waiting thread
 *   - supposing the thread might be revived without the condition being
 *     satisfied (even in case this shouldn't happen) will make your code
 *     more robust
 */
unsigned int thread_wait_usec (const unsigned int usec)
{
	int_state_t int_state;
	struct timer tmr;
	unsigned int usec_left;
	
	save_and_disable_interrupts (int_state);
	
	/* Check that we're not inside a timer handler */
	assert (current_thread->tc_state == TTCS_INACTIVE);
	
	/* Create a timer to revive us */
	timer_init (&tmr, usec, _revive_me, (void *)current_thread->id);
	timer_start (&tmr);
	
	/* Wait for revive() by timer handler or someone else */
	current_thread->rc_state = TRCS_WAITING;
	sched_sleep_thread (current_thread);
	schedule ();
	
	usec_left = timer_usec_left (&tmr);
	timer_destroy (&tmr);
	
	/*
	 * Since the handler is non-preemptible, it had either finished
	 * before the call to timer_destroy(), or didn't run at all
	 */
	
	restore_interrupts (int_state);
	
	return usec_left;
}

/**
 * Suspend the thread for at least \a sec seconds.
 *
 * Equivalent to thread_wait_usec(), but does not return anything.
 * This conforms to the assignment, since a thread waiting in a
 * usleep should never be revived (or it's a bug).
 *
 * Must not be called from a timer handler.
 */
void thread_usleep (const unsigned int usec)
{
	unsigned int usec_remaining;
	
	usec_remaining = thread_wait_usec (usec);
	assert (usec_remaining == 0);
}

/**
 * Give up the rest of current timeslice.
 *
 * This function is meant solely for implementing synchronisation
 * primitives and special purposes. Using it in regular code
 * is considered extremely bad practice.
 */
void thread_yield (void)
{
	int_state_t int_state;
	
	save_and_disable_interrupts (int_state);
	schedule ();
	restore_interrupts (int_state);
}

/**
 * Suspend the current thread.
 *
 * Execution of the current thread is suspended indefinitely,
 * until it is woken up by another thread calling thread_wakeup().
 *
 * This function is intended to be used for synchronising
 * "regular" code. Must not be called from a timer handler.
 */
void thread_suspend (void)
{
	int_state_t int_state;
	
	save_and_disable_interrupts (int_state);

	/* Check that we're not inside a timer handler */
	assert (current_thread->tc_state == TTCS_INACTIVE);
	
	current_thread->rc_state = TRCS_SUSPENDED;
	sched_sleep_thread (current_thread);
	schedule ();
	
	restore_interrupts (int_state);
}

/**
 * Make the current thread wait.
 *
 * Makes the current thread wait for an indefinite amount of time.
 * The thread can be revived by another thread calling
 * thread_revive().
 *
 * As opposed to thread_suspend(), this function is meant to be used
 * to implement synchronisation primitives and must not be used
 * in regular code.
 *
 * Notes:
 * 	- Calling thread_wakeup() on a waiting thread has no effect.
 * 	- must not be called from a timer handler.
 */
void thread_wait (void)
{
	int_state_t int_state;
	
	save_and_disable_interrupts (int_state);
	
	/* Check that we're not inside a timer handler */
	assert (current_thread->tc_state == TTCS_INACTIVE);
	
	current_thread->rc_state = TRCS_WAITING;
	sched_sleep_thread (current_thread);
	schedule ();
	
	restore_interrupts (int_state);
}

/**
 * Bring a thread back from sleep.
 *
 * Internal function used to either "revive" waiting threads
 * (internal use, sync. primitices), or "wakeup" suspended threads
 * (external use, regular kernel code).
 *
 * Restores the regular code state to TRCS_READY. If the thread was
 * sleeping (i.e. no timer handler code was running at the moment),
 * it is also moved to the runnable thread list.
 */
static int _unsleep_from_state (thread_t thr, thread_rc_state_t orig_state)
{
	struct thread_struct *ts;
	int_state_t int_state;

	save_and_disable_interrupts (int_state);
	ts = thread_get_struct (thr);
	if (!ts) {
		restore_interrupts (int_state);
		return EINVAL;
	}

	if(ts->rc_state == orig_state) {
		ts->rc_state = TRCS_READY;
		if (ts->state == TS_SLEEPING)
			sched_unsleep_thread (ts);
	}
		
	restore_interrupts (int_state);

	return EOK;
}


/**
 * Wakes-up a thread suspended by thread_suspend().
 *
 * This function is intended to be used for synchronising
 * "regular" code.
 *
 * Note: Has no effect on waiting threads (in sync. primitives, usleep)
 */
int thread_wakeup (thread_t thr)
{
	return _unsleep_from_state (thr, TRCS_SUSPENDED);
}

/**
 * Revive a waiting thread.
 *
 * As opposed to thread_suspend(), this function is meant to be used
 * to implement synchronisation primitives and must not be used
 * in regular code. Reviving a thread without reason may result
 * in impredictable behavior.
 */
int thread_revive (thread_t thr)
{
	return _unsleep_from_state (thr, TRCS_WAITING);
}


/**
 * Kill a thread.
 * 
 * Stop a thread from being execute and free its
 * stack, thread_struct, id and destroy its pending timers.
 *
 * Notes:
 *   - killing the current thread is forbidden
 *   - killing a thread that might be sleeping on a mutex, semaphore
 *     or condition variable, or holding a lock, is forbidden,
 *     since it would break the protected data structure anyway
 *   - if a thread might have a running timer while being killed,
 *     the timer handler MUST suppose its thread can cease to exist anytime
 *   - killing a thread that is in the middle of thread_sleep(),
 *     thread_usleep() or thread_suspend() is safe
 */
int thread_kill (thread_t thr)
{
	struct thread_struct *ts;
	int_state_t int_state;
	
	save_and_disable_interrupts (int_state);
	ts = thread_get_struct (thr);
	if (!ts) {
		restore_interrupts (int_state);
		return EINVAL;
	}
	
	/* Remove the thread completely */
	_cleanup_thread (ts);	
	
	restore_interrupts (int_state);
	return EOK;
}
