/*
 * PyJaMaS
 *
 * Copyright (c) 2006
 *	Jiri Svoboda
 *	portions from Kalisto
 * see ../README for more details
 *
 * Thread managing routines and data structures.
 */

/** \file
 * Kernel Threads.
 * 
 * See thread.c for details.
 */


#ifndef _THREAD_H_
#define _THREAD_H_

#include "types.h"
#include "sys.h"
#include "list.h"
#include "stackframe.h"

/**
 * Thread status definitions.
 *
 * - TS_DISABLED
 * 	the thread is not prepared to enter the internal structures
 * - TS_READY_TO_RUN
 * 	the thread is active and is waiting for a processor
 * - TS_SLEEPING
 * 	the thread is not waiting for a processor, instead it's waiting
 *      for an event (timout, wakeup, sync. primitive)
 * - TS_RUN
 * 	the thread is active and running
 */
typedef enum {
	TS_DISABLED,
	TS_READY_TO_RUN,
	TS_SLEEPING,
	TS_RUN
} thread_state_t;

/**
 * Thread regular code status.
 *
 * Describes the status of regulare (non-timer) code.
 *
 * - TRCS_READY
 *	regular code is ready to be executed
 * - TRCS_WAITING
 *	regular code is waiting on a sync. primitive
 * - TRCS_SUSPENDED
 *	regular code is waiting for a thread_wakeup()
 */
typedef enum {
	TRCS_READY,
	TRCS_WAITING,
	TRCS_SUSPENDED
} thread_rc_state_t;

/**
 * Thread timer code status.
 *
 * - TTCS_INACTIVE
 *	no timer handler is currently running in the thread
 * - TTCS_ACTIVE
 *	a timer handler is running in the thread right now
 */
typedef enum {
	TTCS_INACTIVE,
	TTCS_ACTIVE
} thread_tc_state_t;


/**
 * Thread identifier.
 *
 * This is a number identifying a thread. This is the safe way
 * to store references to threads. If a thread ceases to exist,
 * the id becomes invalid, which can be safely detected,
 * and remains invalid for a long time. (Until another 2^32 threads
 * have been created).
 */
typedef int32_t thread_t;

/**
 * Thread control structure.
 *
 * "state" describes the current state of the thread. State is useful
 * foir debugging, and to determine whether a thread is waiting on a sync.
 * primitive or is suspended, since there is only one list for all sleeping
 * (that is suspended or waiting) threads.
 *
 * Each synchronization primitive also has a list of all threads waiting
 * on that particular primitive. The current thread is stored in
 * the global variable \c current_thread.
 *
 * We keep the kernel stack as a buffer addressed by the kernel_stack
 * pointer. The "exception_context" variable points to the top of kernel
 * stack where all the general registers are stored after an exception.
 * The "task_switch_context" variable is used by the switch_context
 * function.
 *
 * Thread may be an item of a list.
 */
struct thread_struct {
	thread_t id;
	thread_state_t state;
	thread_rc_state_t rc_state;
	thread_tc_state_t tc_state;
	
	/* The thread's stack and stack pointers */
	struct stack_frame *exception_context;
	struct stack_frame *task_switch_context;
	int *kernel_stack;

	/* Entry of the scheduler's runnable or sleeping list */
	struct list_entry scheduling;
	
	/* ** Join/detach related ** */
	thread_t waited_for_by;
	bool_t is_waited_for;
	bool_t is_finished;
	bool_t is_detached;
	
	/* Entry of the finished_detached_threads list
	   that holds detached threads in need of cleaning-up */
	struct list_entry finished_detached;
	
	/* ** Timer-related ** */
	
	/* List of pending timers owned by this thread. Used for cleaning up. */
	struct list pending_timers;

	/* List of expired timers owned by this thread. */
	struct list expired_timers;
	
	/* ** Synchronization-related data ** */
	struct list_entry sync_queue;
};

/* A pointer to the currently running thread. */
extern struct thread_struct *current_thread;

/*
 * Init
 */

/* Init thread module and create initial thread. */
struct thread_struct *threads_init (void (*start_func) (void *), void *data);

/* Create the service thread. */
void thread_init_service_thread (void);

/*
 * Get
 */
 
/* Return id of the current thread. */			
thread_t thread_get_current (void);

/* Return the thread_struct * associated with an id. */
struct thread_struct *thread_get_struct (thread_t thr_id);

/*
 * Create/kill
 */

/* Create a new runnable thread. */
thread_t thread_create (thread_t *thread_ptr, void (*thread_start) (void *),
			void *data);

/* Kill a thread. */
int thread_kill (thread_t thr);


/*
 * Join/detach
 */

/* Wait for a thread to finish execution. */
int thread_join (thread_t thr);

/* Wait for a thread to finish with timeout. */
int thread_join_timeout (thread_t thr, const unsigned int usec);

/* Detach the current thread. */
int thread_detach (thread_t thr);


/*
 * Timed sleep
 */

/* Make the thread wait for at least \a sec seconds. */
void thread_sleep (const unsigned int sec);

/* Make the thread wait for at least \a sec seconds. */
void thread_usleep (const unsigned int usec);


/*
 * Wait/revive (internal synchronisation)
 */

/* Make the current thread wait. */
void thread_wait (void);

/* Wait for at least \a usec microseconds, or for thread_revive(). */
unsigned int thread_wait_usec (const unsigned int usec);

/* Revive a waiting thread. */
int thread_revive (thread_t thr);

/*
 * Suspend/wakeup (regular code sync.)
 */

/* Suspend the current thread. */
void thread_suspend (void);

/* Wakes-up a thread suspended by thread_suspend(). */
int thread_wakeup (thread_t thr);

/*
 * The rest
 */

/* Give up the rest of current timeslice. */
void thread_yield (void);

/* Switch from current context to context of a given thread. */
void context_switch (struct thread_struct *t);


#endif /* _THREAD_H_ */
