/*
 * PyJaMaS
 *
 * Copyright (c) 2006
 *	Jiri Svoboda
 * see ../README for more details
 *
 */
 
/** \file
 * Kernel Timers.
 *
 * \section timer_lifecycle Lifecycle and handling
 *
 * A timer is always in one of three states: inactive, running or expired.
 * Inactive timers do not participate in kernel structures.
 *
 * Calling timer_init() initialises a timer and makes the calling thread
 * the timers owner. A call to timer_start() changes the state of a timer
 * to running. Running timers are kept in a binary min-heap. A special thread
 * called 'service' periodically runs timer_service_timers(). This function
 * moves all expired timers to the 'expired_timers' queues of their
 * respective owner threads (at that moment the timer state changes
 * to "expired").
 *
 * When a thread gets hold of the CPU, it checks its expired_timers list
 * to see if there is any timer handler to be run. If so, it starts executing
 * the timer handler instead of regular code. When it finishes running
 * all handlers from the list, regular code execution continues.
 * Note that handler execution is naturally preemptible. A thread
 * won't recursively run another timer handler if it is already
 * in the middle of executing one.
 *
 * As we can see, the scheduling of a timer is performed in the context
 * of the thread starting the timer, while the expiration check
 * runs in the context of the 'service' thread. The execution of the handler
 * itself runs in the context of the timer's owner thread.
 * (For more on the service thread, see threads.c).
 *
 * A timer is considered pending if it is either running or expired.
 * (In which case timer_pending() returns true. That means the timer
 * is inactive at the time the handler is running (unless it's restarted).
 * A call to timer_destroy() does not stop a running handler.
 *
 * Since while a timer handler is being executed, its owner thread
 * does not run, the handler execution happens 'instantly' or atomically
 * from their owner thread's point of view.
 *
 * \section timekeeping Timekeeping
 * A very simple timekeeping model is used. The CP0 Count Register is used
 * as a time source. Its period of 2^32 is extended to 2^64,
 * this 64-bit value is returned by _get_current_time_ticks().
 *
 * Internally all time is expressed in Count Register ticks, with
 * 64-bit precision. Microseconds are converted to timer ticks
 * with a fixed factor of TICKS_PER_USEC. As long as this constant
 * is less than 2^32, no timer will have a duration of more than 2^64 ticks
 * and everything will work correctly.
 *
 * Determining whether a timer has expired means comparing its expire time
 * with current time the right way. 'The right way' means
 * 'current_time - expire_time >= 0 ??'. This test, in signed modulo
 * arithmetic, behaves exactly as expected: if current_time and expire_time
 * are less than 2^63-1 apart, they are correctly compared.
 */ 

#include "timer.h"

#include "errno.h"
#include "sys.h"
#include "sched.h"
#include "thread.h"
#include "simdebug.h"
#include "debug.h"
#include "intctl.h"
#include "heap.h"
#include "list.h"

#include "printer.h"

#define TICKS_PER_USEC 8

#define SSD_TIMER_MAGIC 0x71234216

/**
 * Heap of running timers.
 * The first timer to expire is at the top.
 */
static struct heap timer_heap;

/**
 * This corresponds to the "instant" (Counter Register value)
 * at the moment the most recent timer interrupt was issued.
 */
static timer_ticks_t last_interrupt;
static timer_ticks_t next_interrupt;
static timer_ticks_t last_known_time;

static bool_t task_switch_set;

static int _get_current_time_ticks (void);
static void _set_timer_interrupt_at (timer_ticks_t when);
static int _timer_heap_compare (const struct heap_entry *a,
				const struct heap_entry *b);

/**
 * Initialise the timer module.
 */
void timers_init (void) {
	last_interrupt = 0;
	last_known_time = 0;
	task_switch_set = b_false;
	
	heap_init (&timer_heap, _timer_heap_compare);
}

/**
 * Compare expiration time for two timers.
 *
 * Used as a compare function for the timer heap.
 */
static int _timer_heap_compare (const struct heap_entry *a,
				const struct heap_entry *b)
{
	const struct timer *ta, *tb;
	timer_ticks_t diff;
	
	ta = HEAP_DATA_PTR (a, struct timer);
	tb = HEAP_DATA_PTR (b, struct timer);
	
	diff = ta->expires - tb->expires;
	
	if (diff < 0)
		return -1;
	else if (diff > 0)
		return 1;
	else
		return 0;
}


/**
 * Set up a new stask switch after a specified delay.
 *
 * Should be called by the scheduler before each task
 * switch to set time for the next task switch.
 */
void timer_set_task_switch (const unsigned int ticks)
{
	int_state_t int_state;
    
	save_and_disable_interrupts (int_state);
    
	next_interrupt = _get_current_time_ticks() + ticks;
	task_switch_set = b_true;
	
	_set_timer_interrupt_at (next_interrupt);
    
	restore_interrupts (int_state);
}

/**
 * Initialise a timer.
 *
 * Set up a timer with a given duration and handler.
 */
int timer_init (struct timer *tmr, const unsigned int usec,
		void (*handler) (struct timer *, void *), void *data)
{
	if (handler == NULL)
		return EINVAL;

	tmr->duration = usec * TICKS_PER_USEC;
	tmr->handler = handler;
	tmr->data = data;
	tmr->state = ts_inactive;

#ifdef STATIC_STRUCT_DEBUG
	if (!KERNEL_PTR_VALID (tmr))
		panic ("Invalid timer pointer (%p)",tmr);
	tmr->magic = SSD_TIMER_MAGIC;
#endif
	
	list_entry_init (&tmr->expired_timers_entry);
	list_entry_init (&tmr->pending_timers_entry);
	heap_entry_init (&tmr->heap_entry);
    
	return 0;
}

/**
 * Start a timer.
 *
 * The timer state is changed to running and it is inserted into
 * the timer heap.
 */
void timer_start (struct timer *tmr)
{
	int_state_t int_state;

#ifdef STATIC_STRUCT_DEBUG
	if (!KERNEL_PTR_VALID (tmr))
		panic ("Invalid timer pointer (%p)",tmr);
	if (tmr->magic != SSD_TIMER_MAGIC)
		panic ("Invalid timer magic (%x)",tmr->magic);
#endif

	/*
	 * If the timer is pending, stop it first.
	 */
	if (tmr->state != ts_inactive) {
		timer_destroy (tmr);
	}

	/*
	 * Make sure the timer expires no earlier than tmr->duration
	 * from the call to timer_start()
	 */
	tmr->expires = _get_current_time_ticks () + tmr->duration;
	tmr->state = ts_running;
	
	/* Make current thread the owner. */
	tmr->owner = current_thread->id;
    
	save_and_disable_interrupts (int_state);
	heap_insert (&timer_heap, &tmr->heap_entry, (void *)tmr);
	list_insert_back (&current_thread->pending_timers,
			&tmr->pending_timers_entry, (void *)tmr);
	restore_interrupts (int_state);
}

/**
 * Destroy a timer.
 *
 * If the timer is running, it is stopped. Execution of the timer
 * handler is not affected, once it has started.
 *
 * Note: Since there's only one CPU, for a non-preemptible timer,
 * timer_destroy() will never be called at the moment the handler
 * is running.
 */
void timer_destroy (struct timer *tmr)
{
	int_state_t int_state;

	save_and_disable_interrupts (int_state);

#ifdef STATIC_STRUCT_DEBUG
	if (!KERNEL_PTR_VALID (tmr))
		panic ("Invalid timer pointer (%p)",tmr);
	if (tmr->magic != SSD_TIMER_MAGIC)
		panic ("Invalid timer magic (%x)",tmr->magic);
#endif

	switch (tmr->state) {
		case ts_inactive:
			break;
	    
		case ts_running:
			heap_remove (&timer_heap, &tmr->heap_entry);
			list_remove (&tmr->pending_timers_entry);
			break;

		case ts_expired:
			list_remove (&tmr->expired_timers_entry);
			list_remove (&tmr->pending_timers_entry);
			break;
	}
	
	tmr->state = ts_inactive;
	restore_interrupts (int_state);
}

/**
 * Return nonzero if the timer is pending.
 *
 * A timer is pending if it is either running or expired. Otherwise
 * it is not pending, regardless on whether the handler is running
 * or not.
 *
 * The return type is not bool_t just to be sure we are compliant
 * with the specification.
 */
int timer_pending (struct timer *tmr)
{
	int_state_t int_state;
	bool_t is_pending;

	save_and_disable_interrupts (int_state);

#ifdef STATIC_STRUCT_DEBUG
	if (!KERNEL_PTR_VALID (tmr))
		panic ("Invalid timer pointer (%p)",tmr);
	if (tmr->magic != SSD_TIMER_MAGIC)
		panic ("Invalid timer magic (%x)",tmr->magic);
#endif

	is_pending = (tmr->state != ts_inactive);
	
	restore_interrupts (int_state);
	
	return is_pending;
}

/**
 * Return the time remaining to timer expiration.
 *
 * The value zero means the timer has either expired,
 * or will expire soon. In order to determine whether the timer handler
 * has been executed, some side-effect must be introduced into
 * the handler.
 */
unsigned int timer_usec_left (struct timer *tmr)
{
	unsigned int usec_left;
	timer_ticks_t ticks_left;
	int_state_t int_state;
	
	save_and_disable_interrupts (int_state);

#ifdef STATIC_STRUCT_DEBUG
	if (!KERNEL_PTR_VALID (tmr))
		panic ("Invalid timer pointer (%p)",tmr);
	if (tmr->magic != SSD_TIMER_MAGIC)
		panic ("Invalid timer magic (%x)",tmr->magic);
#endif
	
	if (tmr->state != ts_inactive) {
		usec_left = 0;
	} else {
		ticks_left = tmr->expires - _get_current_time_ticks();
		if (ticks_left < 0)
			ticks_left = 0;
		usec_left = ticks_left / TICKS_PER_USEC;
	}
	restore_interrupts (int_state);
	
	return usec_left;
}

/**
 * Destroy all pending timers belonging to a thread.
 *
 * Note: Running handlers, if any, will continue their execution.
 */
void timer_destroy_pending_timers (struct thread_struct *ts)
{
	struct list_entry *tmr_entry, *next_entry;
	struct timer *tmr;	
	int_state_t int_state;
	
	save_and_disable_interrupts (int_state);

	tmr_entry = list_first (&ts->pending_timers);
	while (!list_is_head (&ts->pending_timers, tmr_entry)) {
	
		next_entry = list_next (tmr_entry);
		tmr = LIST_DATA_PTR (tmr_entry, struct timer);
		
		timer_destroy (tmr);
		
		tmr_entry = next_entry;
	}
	
	restore_interrupts (int_state);
}

/**
 * Move all expired timers from timer heap to expired list.
 */
static void _enlist_expired_timers (void)
{
	struct heap_entry *e_earliest;
	struct timer *t_earliest;
	struct thread_struct *owner_ts;
	timer_ticks_t now;
	int_state_t int_state;
	
	save_and_disable_interrupts (int_state);
	
	now = _get_current_time_ticks ();
    
	while (1) {
		e_earliest = heap_get_top (&timer_heap);
		if (!e_earliest)
			break;

		t_earliest = HEAP_DATA_PTR (e_earliest, struct timer);
		
		/*
		 * Works OK as long as the two values are less
		 * than 2^63 apart.
		 */
		if ( !(t_earliest->expires - now <= 0) )
			break;
	
		/* The timer expired */
			
		/*
		 * Get the owner's thread struct.
		 */
		owner_ts = thread_get_struct(t_earliest->owner);
		assert(owner_ts != NULL);
		
		/*
		 * Insert to the owner's expired list.
		 */
		list_insert_back (&owner_ts->expired_timers,
			&t_earliest->expired_timers_entry, (void *)t_earliest);
		t_earliest->state = ts_expired;
		
		/* 
		 * Check whether the thread was not sleeping. If it was,
		 * unsleep it.
		 */
		if (owner_ts->state == TS_SLEEPING)
			sched_unsleep_thread (owner_ts);
		
		heap_remove_top (&timer_heap);
	}

	restore_interrupts (int_state);
}

/**
 * Return the current timing value in timer ticks.
 *
 * Takes the value of the cp0_count register and extends it
 * to 64 bits, with the assumption that less than CP0_COUNT_RANGE
 * (i.e. 2^32) ticks have passed since the last timer interrupt.
 */
static int _get_current_time_ticks (void)
{
	timer_ticks_t current_time;
	
	current_time = (last_known_time & ~(CP0_COUNT_MASK))
	  | read_cp0_count ();

	if (current_time - last_known_time < 0)
		current_time += CP0_COUNT_RANGE;

	last_known_time = current_time;

	return current_time;
}

/**
 * Set compare register to interrupt us at a given time.
 */
static void _set_timer_interrupt_at (timer_ticks_t when)
{
	/*
	 * Set a new time till timer interrupt
	 *
	 * TODO: is this the right place?? (it's HW-dependent)
	 * TODO: check range of when
	 */
	write_cp0_compare (when);
}

/**
 * Run all expired timer handlers for the current thread
 *
 * This is executed in a thread's 'timer' context of execution,
 * with interrupts enabled.
 */
void timer_run_handlers (void)
{
	struct list_entry *tmr_entry;
	struct timer *tmr;

	void (*timer_handler) (struct timer *, void *);
	void *timer_data;
	int_state_t int_state;
	
	save_and_disable_interrupts (int_state);

	while (!list_is_empty (&current_thread->expired_timers)) {
		/* Fetch the timer */
		tmr_entry = list_first (&current_thread->expired_timers);
		tmr = LIST_DATA_PTR (tmr_entry, struct timer);
		
		timer_handler = tmr->handler;
		timer_data = tmr->data;
		
		/* Remove it from lists / change state to inactive */
		list_remove (tmr_entry);
		list_remove (&tmr->pending_timers_entry);

		tmr->state = ts_inactive;
		
		/*
		 * Run the handler - with interrupts enabled
	         */
		restore_interrupts (int_state);
		(*timer_handler) (tmr, timer_data);
		save_and_disable_interrupts (int_state);
	}
	
	restore_interrupts (int_state);
}

/**
 * Enlist expired timers and run one timer handler.
 *
 * Returns b_true if there is more work to be done, b_false otherwise.
 * This is called periodically by the service thread, see thread.c.
 */
void timer_service_timers (void)
{
	_enlist_expired_timers ();
}


/**
 * Service a Timer Interrupt.
 *
 * Called from the time interrupt handling routine.
 *
 * Update time counters and request another task switch time to be set,
 * then run the scheduler.
 *
 * Runs with interrupts disabled.
 */
void timer_interrupt (void)
{    

	last_interrupt = next_interrupt;
	task_switch_set = b_false;
	
	/*
	 * Before switching tasks, the scheduler will call
	 * timer_set_task_switch() to clear the interrupt
	 * and prepare the next one.
	 */
	schedule ();
}
