/*
 * PyJaMaS
 *
 * Copyright (c) 2006
 *	Jiri Svoboda
 * see ../README for more details
 *
 */
 
/** \file
 * Kernel Timers.
 * 
 * See timer.c for details.
 */

#ifndef _TIMER_H_
#define _TIMER_H_

#include "types.h"
#include "heap.h"
#include "list.h"
#include "thread.h"

/**
 * Current state of a timer.
 *
 *   - \c ts_inactive Not started yet, or already serviced.
 *   - \c ts_running Started, but expiration not detected yet.
 *   - \c ts_expired Expiration detected, but handler not started yet.
 */
typedef enum {
	ts_inactive,
	ts_running,
	ts_expired
} timer_state_t;

/**
 * For storing and comparing timer values.
 *
 * For timers to work, we need max(timer duration in usec)*TICKS_PER_USEC
 * to fit into timer_ticks_t. This should hold as long as
 * TICKS_PER_USEC < 2^31.
 */
typedef int64_t timer_ticks_t;

/**
 * Timer Structure.
 */
struct timer {
	/** Current timer state */
	timer_state_t state;
	
	/** Timer duration in time ticks */
	timer_ticks_t duration;
	
	/** When the timer is going to expire */
	timer_ticks_t expires;
	
	/** Whether interrupts are enabled during handler execution */
	bool_t preemptible_handler;
	
	/** Handler function */
	void (*handler) (struct timer *, void *);
	
	/** Data to pass to the handler function */
	void *data;
	
	/** For participating in the running timers heap */
	struct heap_entry heap_entry;
	
	/** For participating in the expired timers list */
	struct list_entry expired_timers_entry;
	
	/**
	 * Entry of owner thread's pending timer list.
	 * Used to distroy all timers of a thread being killed
	 */
	struct list_entry pending_timers_entry;
	
	/**
	 * Thread owning this timer.
	 * Needed to unsleep the thread on timer expiry.
	 */
	thread_t owner;

#ifdef STATIC_STRUCT_DEBUG
	unsigned magic;
#endif
};


/* Initialise the timer module. */
void timers_init (void);

/* Initialise a timer. */
int timer_init (struct timer *tmr, const unsigned int usec,
		void (*handler) (struct timer *, void *), void *data);

/* Initialise a timer (internal). */
void timer_start (struct timer *tmr);

/* Destroy a timer. */
void timer_destroy (struct timer *tmr);

/* Return nonzero if the timer is pending. */
int timer_pending (struct timer *tmr);

/* Return the time remaining to timer expiration. */
unsigned int timer_usec_left (struct timer *tmr);

/* Destroy all pending timers belonging to a thread. */
void timer_destroy_pending_timers (struct thread_struct *ts);

/* Set up a new stask switch after a specified delay. */
void timer_set_task_switch (const unsigned int ticks);

/* Move all expired timers from timer heap to expired list. */
void timer_interrupt (void);

/* Run all expired timer handlers for the current thread */
void timer_run_handlers (void);

/* Enlist expired timers. */
void timer_service_timers (void);

#endif /* _TIMER_H_ */
